/**
 * Generate api guide.
 * @memberof module:apiguide/lib
 * @function apiguide
 * @param {string} filenames - Source file names.
 * @param {object} [options] - APIGuide options.
 * @param {string} [options.verbose] - Verbose or not.
 * @param {string} [options.tutorials] - Directory path for tutorials.
 * @param {string} [options.destination] - Destination directory path.
 * @param {string} [options.templates] - Markup options.
 * @param {string} [options.templates.systemName] - Name of theme.
 * @param {string} [options.templates.footer] - Footer html.
 * @param {string} [options.templates.copyright] - Copyright string.
 * @param {string} [options.templates.color]  - Theme color.
 * @param {function} callback - Callback when done.
 */

"use strict";

var async = require('async'),
    path = require('path'),
    argx = require('argx'),
    fs = require('fs'),
    _generateTheme = require('./_generate_theme'),
    _execJsdoc = require('./_exec_jsdoc');


var TEMPLATE_DIR = path.resolve(__dirname, '../assets/jsdoc_template');

/** @lends apiguide */
function apiguide(filenames, options, callback) {
    var args = argx(arguments);
    callback = args.pop('function');
    options = args.pop('object') || {};
    filenames = args.remain().reduce(function (a, b) {
        return a.concat(b);
    }, []);
    var destination = options.destination;
    if (!destination) {
        callback(new Error('options.destination is required.'));
        return;
    }
    var favicon = options.favicon || "favicon.png";

    var configure = path.resolve(process.cwd(), '.apiguide.jsdoc_config.json');

    var templates = options.templates || {};
    async.series([
        function (callback) {
            _generateTheme(destination, {
                color: templates.color || "#665FF5"
            }, callback);
        },
        function (callback) {
            fs.writeFile(configure, JSON.stringify({
                tags: {
                    "allowUnknownTags": true
                },
                plugins: ["plugins/markdown"],
                templates: {
                    cleverLinks: false,
                    monospaceLinks: false,
                    outputSourceFiles: true,
                    theme: 'dynamic',
                    systemName: templates.systemName,
                    footer: [
                        "<script>",
                        "var h=document.head;h.innerHTML=h.innerHTML+'<link rel=\"shortcut icon\" href=\"" + favicon + "\"/>';",
                        "</script>",
                        templates.footer
                    ].join(''),
                    copyright: templates.copyright,
                    navType: "vertical",
                    linenums: true,
                    collapseSymbols: false,
                    dateFormat: "YYYY-MM-DD",
                    inverseNav: false
                },
                markdown: {
                    parser: "gfm",
                    hardwrap: true
                }
            }), callback);
        },
        function (callback) {
            _execJsdoc(filenames, {
                verbose: !!options.verbose,
                template: TEMPLATE_DIR,
                destination: destination,
                configure: configure
            }, callback);
        },
        function (callback) {
            fs.unlink(configure, callback);
        }
    ], callback);
}

module.exports = apiguide;
